# -*- coding: utf-8 -*-
"""
Created on Sun Jan 19 09:07:43 2020

@author: Jean
"""

from dli.table import TableSchema
from dli.column import Column
from dli.exception import DliException
from dli.dli_client import DliClient
import time
import os
import sys
import pandas as pd
from io import BytesIO
from obs import *
from Orange.data.pandas_compat import table_from_frame
from Orange.data import DiscreteVariable
from Orange.data.table import Table
from Orange.data import Domain
import numpy as np


def init_obs_client():
    AK = '***************'
    SK = '******************************'
    server = 'https://obs.cn-south-1.myhuaweicloud.com'
    # Constructs a obs client instance with your account for accessing OBS
    obsClient = ObsClient(access_key_id=AK, secret_access_key=SK, server=server)
    return obsClient


def init_aksk_dli_client():
    auth_mode = 'aksk'
    region = 'cn-south-1'
    project_id = '076592e9080010e12ff7c01682201f7b'
    ak = '***********************'
    sk = '*********************************'
    dli_client = DliClient(auth_mode=auth_mode, region=region, project_id=project_id,ak=ak, sk=sk)
    return dli_client

def list_all_queues(dli_client):
    try:
        queues = dli_client.list_queues()
    except DliException as e:
        print(e)
        return

    for queue in queues:
        print(queue.name)


def list_all_dbs(dli_client):
    try:
        dbs = dli_client.list_databases()
    except DliException as e:
        print(e)
        return

    for db in dbs:
        print(db)


def list_all_tbls(dli_client, db_name):
    try:
        tbls = dli_client.list_tables(db_name, with_detail=True)
    except DliException as e:
        print(e)
        return

    for tbl in tbls:
        print(tbl.name)


def create_dli_tbl(dli_client, db_name, tbl_name, cols, sort_cols=None):
    tbl_schema = TableSchema(tbl_name, cols, sort_cols)
    try:
        table = dli_client.create_dli_table(db_name, tbl_schema)
    except DliException as e:
        print(e)
        return

    print(table)


def delete_tbl(dli_client, db_name,tbl_name):
    try:
        dli_client.delete_table(db_name, tbl_name)
    except DliException as e:
        print(e)
        return


def run_sql(dli_client, db_name, queue_name, sql):
    # execute SQL
    try:
        sql_job = dli_client.execute_sql(sql, db_name)
        result_set = sql_job.get_result()
    except DliException as e:
        print(e)
        return

    if result_set.row_count == 0:
        return None

    return result_set

def insert_sql(dli_client, db_name, queue_name, sql):
    # execute SQL
    try:
        sql_job = dli_client.execute_sql(sql, db_name)
    except DliException as e:
        print(e)
        return
    
    print(sql_job.job_type)
    print(sql_job.job_status)


db_name = 'geshui'
queue_name = 'default'
bucketName = 'gsmx'

dli_client = init_aksk_dli_client()
obsClient = init_obs_client()

# 测试DLI Python API
t1 = time.time()
y = [] ; c=[] 
sql = "select date_format(ssqq, 'y') as y, count(*) as c from gsmx group by date_format(ssqq, 'y');"
result_set = run_sql(dli_client, db_name, queue_name, sql)
i= 0
print("loading data......")
for row in result_set:
    cols = row.columns
    y.append(cols[0])
    c.append(cols[1])
    i = i+1
    print(i," : ",cols)
t2 = time.time()
print("time: ",t2-t1)

# 太慢，改用OBS下载，每次1000条，还是限量版，比去年每次100条有改进。
t1 = time.time()
djxh = [] ; y=[] ; pjgz = []
sql = "select djxh,cast(y as int) as y, pjgz from dwrstj3 order by djxh,y;"
result_set = run_sql(dli_client, db_name, queue_name, sql)
i= 0
print("loading data......")
for row in result_set:
    cols = row.columns
    djxh.append(cols[0])
    y.append(cols[1])
    pjgz.append(cols[2])
    i = i+1
    print(i," : ",cols)
t2 = time.time()
print("time: ",t2-t1)


# # 从 list建立 pandas data frame
# data = pd.DataFrame({"djxh":djxh,"y":y,"pjgz":pjgz})
# print(data[0:5])

def sql_to_dataframe(dli_client,obsClient,table,cols,types,sql):
    # Select语句先输出CSV数据文件到OBS
    insert_sql(dli_client, db_name, queue_name, sql)
    # 再从OBS下载CSV格式数据
    resp = obsClient.listObjects(bucketName,"table/"+table)
    data = None
    for content in resp.body.contents:
        if content.key == "table/"+table+"/_SUCCESS":
            continue
        print('Downloading :',content.key)
        resp = obsClient.getObject(bucketName, content.key, loadStreamInMemory=True)
        fileBytes = BytesIO( resp.body.buffer )
        df = pd.read_csv(fileBytes,header=None,names=cols,dtype=types)
        if data is None:
            data = df
        else:
            data = pd.concat([data,df],ignore_index=True)
    # 最后删除OBS CSV中间文件   
    resp = obsClient.listObjects(bucketName,"table/"+table)
    for content in resp.body.contents:
        obsClient.deleteObject(bucketName, content.key)
    obsClient.deleteObject(bucketName, "table/"+table)   
    return data

table = "dwrstj3"
cols = ["djxh","y","pjgz"]; types = {"djxh":str,"y":int,"pjgz":float}
sql = "INSERT OVERWRITE DIRECTORY 'obs://gsmx/table/dwrstj3' USING csv \
            select djxh,cast(y as int) as y, pjgz from dwrstj3 order by djxh,y;"
t1 = time.time()
data = sql_to_dataframe(dli_client,obsClient,table,cols,types,sql)
# data["y"] = data["y"] - 2000
t2 = time.time()
print(t2-t1)  
print(data.describe())
print(data[0:10])

# 山脊型连续下降模式识别函数,返回风险年度，下降百分比，下降额度3个指标
# df 个税年平均工资按登记序号+年 汇总数据，pandas data frame
# steps:要向前检测的点数，>=1
# target:目标点是当前点的百分比, 0<=target<1
# wave:中间的点反弹形成波峰允许的反弹系数范围，wave>1
def isRisk(df,steps=3,target=0.8,wave=1.2):
    ymin = 0 ; ymax = 0
    length = df.y.count()
    for i in range(length):
        if df.iloc[i,2] > df.iloc[ymax,2]:
            ymax=i
        if df.iloc[i,2] < df.iloc[ymin,2]:
            ymin=i               
    risk = 0 ; perc = 0; amount = 0
    if df.iloc[ymax,2]<=0 or length<steps:
        return (risk,perc,amount)
    for i in range(length-steps):
        ok=0
        if df.iloc[i,2]>0 and df.iloc[i+steps,2] / df.iloc[i,2]<= target:
            for j in range(steps):
                if steps ==1:
                    w = df.iloc[i+j,2]
                else:
                    w = df.iloc[i+j,2]*wave
                if df.iloc[i+j+1,2]> w:
                    break
                else:
                    ok = ok+1
        if ok==steps:
            risk = df.iloc[i,1]
            perc = round((1-df.iloc[i+steps,2] / df.iloc[i,2])*1000)/10
            amount = round((df.iloc[i,2] - df.iloc[i+steps,2])*100)/100
            # print(risk,perc,amount)
            break
    return (risk,perc,amount)


# 测试单个登记序号汇总数据的风险识别
djxhs = data["djxh"]
#djxhs2 = list(set(djxhs.values))
index = djxhs.str.contains("10124404000003626510")
df = data[index]
df = df.sort_values(by="y" , ascending=True)
print(df)
steps =3
target = 0.8
wave = 1.2  
risk,perc,amount = isRisk(df,steps,target,wave)      
print(risk,perc,amount)

# 识别所有登记序号的风险
t1 = time.time()        
xhs = []; risks =[]; percs =[]; amounts =[]
# 用pandas groupby() 功能按登记序号分组识别风险
dgroup =data.groupby("djxh")
# 逐组识别风险
i=0
for dg in dgroup:
    i = i+1
    djxh = dg[0] ; df = dg[1]
    df = df.sort_values(by="y" , ascending=True)
    risk, perc,amount = isRisk(df,3,0.8,1.2)
    xhs.append(djxh); risks.append(risk); percs.append(perc); amounts.append(amount)
    print(i,djxh,risk,perc,amount)

# 风险识别结果转换成 pandas data frame
riskTags = pd.DataFrame({"djxh":xhs,"risk":risks,"perc":percs,"amount":amounts})
t2 = time.time()
print(t2-t1)
# 输出风险识别结果统计    
print(riskTags.describe())


# pandas data frame数据上传至表
def upload_dataframe(dli_client,obsClient,data,table):
    tableFile = table+".csv"
    key = "table/"+tableFile
    # data frame数据导出至 CSV
    data.to_csv(tableFile,index=None)
    # CSV上传至 OBS
    resp = obsClient.uploadFile(bucketName,key,tableFile)
    print(resp)
    # 从OBS导入数据至表
    sql = "load data inpath 'obs://gsmx/"+key+\
          "' into table "+table+" options('DATA_TYPE'='CSV','HEADER'='true');"
    print(sql)
    insert_sql(dli_client, db_name, queue_name, sql)
    #删除临时文件
    obsClient.deleteObject(bucketName, key)   
    os.remove(tableFile)
 
t1 = time.time()
# 建立DLI表上传风险识别结果
cols = [
    Column('djxh', 'string'),
    Column('risk', 'int'),
    Column('perc', 'double'),
    Column('amount', 'double')
]
sort_cols = ['djxh']
delete_tbl(dli_client, db_name,"dwtj2")
create_dli_tbl(dli_client, db_name, "dwtj2", cols, sort_cols)
# 上传数据到该表
upload_dataframe(dli_client, obsClient, riskTags, "dwtj2")
t2 = time.time()
print(t2-t1)  


# 云端合并风险标签与风险特征库
t1 = time.time()
sql = "create table dwtj3 as \
    select a.*,b.risk,b.perc,b.amount from dwtj a, dwtj2 b where a.djxh=b.djxh;"
delete_tbl(dli_client, db_name,"dwtj3")
insert_sql(dli_client, db_name, queue_name, sql)
# 合并后的数据重新下载到客户端进行机器学习等分析
table = "dwtj3" ; cols=["sdjxh"] 
schema = dli_client.get_table_schema(db_name,table)
for column in schema.columns:
    cols.append(column.name)
cols.append("isRisk")
sql = "INSERT OVERWRITE DIRECTORY 'obs://gsmx/table/dwtj3' USING csv \
            select concat('D',a.djxh) as sdjxh, a.*,\
                (case when a.risk>0 then 1 else 0 end) as isRisk \
                from dwtj3 a order by a.djxh;"
riskData = sql_to_dataframe(dli_client,obsClient,table,cols,None,sql)
t2 = time.time() ;
print(t2-t1)
  
# 提取风险标签列，删除不用的列等预处理
isRisk=riskData["isRisk"]
riskData.drop(["djxh","kzztdjlx_dm","djzclx_dm","hy_dm",\
                "risk","perc","amount","isRisk"],axis=1, inplace=True) 
riskData.rename(columns = {"sdjxh": "djxh"}, inplace=True)
riskData["djxh"] = riskData["djxh"].str.replace("D","")
# 生成 Orange格式 data table，添加标签列isRisk为目标列，调用并传送到Orange
otable = table_from_frame(riskData)
target = DiscreteVariable("isRisk",["False","True"])
domain= Domain(otable.domain.attributes,target,otable.domain.metas)
otable2 = Table.from_numpy(domain,otable.X,Y=np.array(isRisk),\
                           metas=otable.metas,W=otable.W)
otable2.name = "个税汇总" ; del otable
from Orange.canvas import __main__ as OrangeMain
sys.exit(OrangeMain.main(datasets = [otable2], logfile="orange.log"))

riskData.info()
